#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

FILE * infile ;
int nt , nc ;
char taxname[10] ;
char tmpname[50] ; 
#define TOLERANCE 0.00001
#define MAXCHAR 12000
#define MAXSWAPS 380 // This is the max number of TBR rearrangs (on 10 taxa)
signed char matrix[20][MAXCHAR] ; 


int nopooprun = 0 ; 
signed long int pathsfrom = -1 ; 
int testgroup[25] , testgroupsz[25] ;
int numtestgroups = 0 ;
unsigned long int testmat[20] ; 

double mptlen = 10000000 ;
double lptlen = 0 ;

#define NEP 2.7182818284590452353602874713527

double damaxprob = 0 , daminprob = 10 ;

typedef struct {
    char tree [ 20 ] ;
    unsigned long int key ;
    unsigned short int lokey ; 
    double len ;
    int tbrlist[MAXSWAPS] ;
    int nmoves ; 
    char isopt ;
    void * to ;
    double prob ; } Treetyp ; 
Treetyp * trees ;
Treetyp ** treelist , ** ordlist ; 
long int ntrees , nodz ;
signed char anx[20] , sis[20] , lef[20] , rig[20] , nodfork[20] ;
signed char hsis[20] , hlef[20] , hrig[20] ;
double blen[20] ; 
char * anxpt ;
unsigned long int curkey ;
unsigned short int curlokey ;

unsigned long int * bitstring ; 
int biz ; 
int deldupes = 1 ;
unsigned long int nummpts = 0 , numnonmpts = 0 ;

void addtolist ( Treetyp * , double ) ; 
double getdalik ( FILE * ) ; 
double findmylik ( void ) ; 

int use_spr = 0 ;
int use_nni = 0 ;

char testpartit[25][10] , partitsize[25] , partitat[25] ;
int numtestpartits = 0 ;

void addtestgroup ( char * txt )
{
    int a ;
    char * cp = txt ;
    int slashisat = -1 , isapartit = 0 ;
    while ( * cp ) if ( * cp ++ == '/' ) isapartit = 1 ; 
    if ( isapartit ) {
       partitsize [ numtestpartits ] = partitat [ numtestpartits ] = 0 ; 
       cp = txt - 1 ;
       isapartit = 0 ; 
       while ( * ++ cp ) {
           if ( * cp == '/' ) {
               if ( partitsize [ numtestpartits ] < 2 ) break ;
               partitat [ numtestpartits ] = slashisat = partitsize [ numtestpartits ] ;
               continue ; }
           for ( a = 0 ; a < nt ; ++ a )
             if ( tolower ( taxname[a] ) == tolower ( * cp ) ) 
                testpartit [ numtestpartits ] [ partitsize [ numtestpartits ] ++ ] = a ; }
       if ( partitsize [ numtestpartits ] < 4 ||  partitsize [ numtestpartits ] - slashisat < 2 ) {
           fprintf ( stderr , "\nOOPS!! wrong partition definition: %s \nDefinition must include 2 taxa (or more) on each side.\n" , txt ) ;
           exit ( 0 ) ; }
       ++ numtestpartits ; }
    else { 
       cp = txt - 1 ;
       testgroup [ numtestgroups ] = testgroupsz [ numtestgroups ] = 0 ;
       while ( * ++ cp ) 
           for ( a = 0 ; a < nt ; ++ a )
             if ( tolower ( taxname[a] ) == tolower ( * cp ) ) {
                ++ testgroupsz [ numtestgroups ] ; 
                if ( testgroupsz [ numtestgroups ] == nt - 1 ) {
                    fprintf ( stderr , "\nOOPS!! Error in group definition: %s\nDefinition can include at most T-2 taxa.\n" , cp ) ;
                    exit ( 0 ) ; }
                testgroup [ numtestgroups ] |= 1 << a ; }
       ++ numtestgroups ; } 
}    

void showlist ( void )
{
    int a ;
    int showed = 0 ; 
    Treetyp * entry ; 
    for ( a = 0 ; a < 65536 ; ++ a ) {
        entry = ( Treetyp * ) treelist [ a ] ;
        if ( entry == NULL ) continue ;
        fprintf ( stdout , "KEY %i: " , a ) ;
        ++ showed ; 
        fprintf ( stdout , "%i" , entry - trees ) ;
        entry = ( Treetyp * ) entry -> to ; 
        while ( entry != NULL ) {
            ++ showed ; 
            fprintf ( stdout , ", %i" , entry - trees ) ;
            entry = ( Treetyp * ) entry -> to ; }
        fprintf ( stdout , "\n" ) ; }
    fprintf ( stdout , "SHOWED %i ENTRIES\n" , showed ) ; 
}

void readdata ( char * nam )
{
    char * nampt ;
    char state ;
    int a , b ;
    if ( nam[0] == '*' ) { nopooprun = 1 ; ++ nam ; }
    infile = fopen ( nam , "rb" ) ;
    if ( infile == NULL ) {
        fprintf ( stderr , "Can't open %s" , nam ) ;
        exit ( 0 ) ; }
    fscanf ( infile , " %s" , tmpname ) ;
    if ( strcmp ( tmpname , "xread" ) ) {
        fprintf ( stderr , "String \"xread\" must appear first in data set" ) ;
        exit ( 0 ) ; }
    fscanf ( infile , " %i %i" , &nc , &nt ) ;
    if ( nt > 10 ) {
        fprintf ( stderr , "Cannot run more than 10 taxa" ) ;
        exit ( 0 ) ; }
    if ( nc > MAXCHAR) {
        fprintf ( stderr , "Cannot run more than %i characters" , MAXCHAR ) ;
        exit ( 0 ) ; }
    for ( a = 0 ; a < nt ; ++ a ) {
        nampt = tmpname ; // 
        fscanf ( infile , " %s" , nampt ) ;
        if ( strlen ( nampt ) > 1 ) {
            fprintf ( stderr , "Taxon names must be single letters!" ) ;
            exit ( 0 ) ; }
        taxname[a] = * nampt ; 
        for ( b = 0 ; b < nc ; ++ b ) {
            fscanf ( infile , " %c" , &state ) ;
            if ( state == '?' || state == '-' ) matrix[a][b] = 255 ;
            else {

if ( toupper ( state ) == 'A' || 
     toupper ( state ) == 'G' || 
     toupper ( state ) == 'C' || 
     toupper ( state ) == 'T' ) {
             if ( toupper(state) == 'A' ) matrix[a][b] = 1 ;
        else if ( toupper(state) == 'G' ) matrix[a][b] = 2 ;
        else if ( toupper(state) == 'C' ) matrix[a][b] = 4 ;
        else if ( toupper(state) == 'T' ) matrix[a][b] = 8 ; }
    else {
                if ( state > '3' ) {
                   fprintf ( stderr , "Cannot run more than 3 states (0-3)" ) ;
                   exit ( 0 ) ; }
            matrix[a][b] = 1 << ( state - '0' ) ; }}}}
    fclose ( infile ) ; 
    return ;
}        

char * bchlist , * bast ;
static int listhas, htuz, agre, totax;

void callyou ( void )
{
    int a, b, c, x, y ; 
    htuz++;
    y = agre ; 
    for ( a = listhas; a--; ) {
       b = bchlist[a];
       x = bast[b];
       bast[y] = bast[b] = htuz;
       bast [ htuz ] = x;
       bchlist[listhas++] = y;
       bchlist[listhas++] = htuz; 
       if ( ++agre < totax ) callyou();
       else { 
              bast [ nt ] = nt+totax-1;
              for( c = 0 ; c < nodz ; ++ c )
                 trees[ ntrees ].tree [c] = bast[c];
              ++ ntrees ; } 
       listhas -= 2;
       --agre; 
       bast [ b ] = x; } 
    htuz--;
} 

void generalltrees ( void )
{
    int a , b , c ;
    char this[ 20 ] ;
    char listspace [ 20 ] ;
    FILE * poo ; 
    a = 4 ; 
    ntrees = 3 ; 
    while ( a ++ < nt ) 
        ntrees *= ( 2 * a ) - 5 ;
    fprintf ( stderr , "Generating %i trees..." , ntrees ) ;
    trees = ( Treetyp * ) malloc ( ntrees * sizeof ( Treetyp ) ) ;
    if ( trees == NULL ) {
        fprintf ( stderr , "\nOOPS...\nCannot allocate needed memory\n" ) ;
        exit ( 0 ) ; }
    biz = ( ntrees + 31 ) / 32 ;
    bitstring = malloc ( biz * sizeof ( unsigned long int ) ) ; 
    treelist = ( Treetyp ** ) malloc ( 65536 * sizeof ( Treetyp * ) ) ;
    if ( treelist == NULL ) {
        fprintf ( stderr , "\nOOPS...\nCannot allocate needed memory\n" ) ;
        exit ( 0 ) ; }
    if ( nopooprun ) {
        nodz = 2*nt - 1 ; 
        read_all_trees () ;
        return ; }
    poo = fopen ( "poofile.nex" , "w" ) ;
    fprintf ( poo , "\
#NEXUS\n\
begin data ;\n\
dimensions nchar=%i ntax=%i;\n\
format datatype=DNA missing=?;\n\
matrix\n" , nc , nt ) ;

for ( a = 0 ; a < nt ; ++ a ) {
    fprintf ( poo , "\n%c " , taxname[a] ) ;
    for ( b = 0 ; b < nc ; ++ b ) {
        c = matrix [ a ] [ b ] ;
        if ( c == -1 ) fprintf ( poo , "?" ) ;
        else if ( c == 1 ) fprintf ( poo , "A" ) ; 
        else if ( c == 2 ) fprintf ( poo , "C" ) ; 
        else if ( c == 4 ) fprintf ( poo , "G" ) ; 
        else if ( c == 8 ) fprintf ( poo , "T" ) ; }}
fprintf ( poo , "\n;end;\n" ) ;
fprintf ( poo , "begin paup;\nset crit=lik;\nlset nst=1 trat=1 rates=eq basef=eq;\nset nowarnreset nomonitor;\n" ) ;
fprintf ( poo , "generate all;\nlog file = allliks.tmp;\nlsco;\nsavetree replace;\nquit nowarn;\n" ) ; 
fclose ( poo ) ; 
system ( "del poofile.tre" ) ; 
system ( "del allliks.tmp" ) ; 
system ( "paup poofile.nex" ) ;
nodz = 2*nt - 1 ; 
read_all_trees () ;
return ; 
} 

void hashmksis ( char * an )
{ signed char list [ 20 ] ; 
  signed char * x = list , * y = list + 1;
  signed char * last , nd ; 
  int a , i;
  last = list ; 
  for ( a = nodz ; a -- ; ) 
      hsis [ a ] = hlef [ a ] = list [ a ] = -1 ;
  for ( a = nt ; a -- ; ) {
      i = an [ nd = a ] ; 
      if ( !i ) continue ; 
      while ( i != nt && hlef [ i ] < 0 ) {
                    last [ i ] = hlef [ i ] = nd ;
                    nd = i ;
                    i = an [ i ] ; }
      if ( hlef [ i ] < 0 ) hlef [ i ] = nd ; // i.e. i == nt
      if ( last [ i ] >= 0 ) hsis [ last [ i ] ] = nd ;
      last [ i ] = nd ; }
   return ;
} 

void reset_nodz ( char * an )
{
  signed char bu [ 20 ] ; 
  signed char tbu [ 20 ] ;
  int a, b, atno, lg ; 
  for ( a = nodz ; a-- > nt ; ) tbu [ a ] = -1;
  for ( a = nt ; a-- ; ) tbu [ a ] = a ; 
  atno = nt ;
  lg = an [ nt ] ; 
  tbu [ nt ] = nt ;
  tbu [ lg ] = lg ; 
  for ( a = 0 ; a < nt ; ++a) {
      b = an [ a ] ;
      if ( !b ) continue ; 
      while ( tbu [ b ] < 0 ) {
          tbu [ b ] = ++ atno ; 
          b = an [ b ] ; }}
  for ( a = lg ; a-- ; ) 
      bu [ tbu [ a ] ] = tbu [ an [ a ] ] ;
  for ( a = lg ; a-- ; ) an [ a ] = bu [ a ] ;
  an [ nt ] = ++ atno ; 
  return ; 
}

unsigned long int topnod; 
short int bchbel [ 20 ] ;
long int needbitz ;

static unsigned long int bm[] = {
       0,1,3,7,15,31,63,127,255,511,1023,2047,4095,8201 } ;
       
void hashtree ( char * treept )     // Make a unique 32-bit tree-identifier... 
{ unsigned long int a, b, n, jamp, at, low;
  unsigned long int bitz, value, ovr, anxwas; 
  hashmksis( treept ) ;
  for( a = 1 ; a < nodz ; ++ a ) bchbel [ a ] = -1 ;
  bchbel [ a = 0 ] = 0;
  while( a != nt ) { a = treept [ a ] ; bchbel [ a ] = 0 ; }
  a = 1 ; while ( bchbel [ a ] < 0 ) { bchbel [ a ] = 1 ; a = treept [ a ] ; }
  bchbel [ a ] = 2 ; while( a != nt ) { bchbel [ a ] = 2 ; a = treept [ a ] ; }
  bchbel [ nt ] = jamp = bitz = 2;
  needbitz = 0; 
  for ( a = nt ; a < nodz ; ++ a ) hrig [ a ] = 0 ;
  a = 0 ; while ( a != nt ) { a = treept [ a ] ; ++ hrig [ a ] ; }
  a = treept [ 1 ] ; while( ! hrig [ a ] ) { ++ hrig [ a ] ; a = treept [ a ] ; }
  ++ hrig [ a ] ; 
  curkey = 0 ; 
  ovr = 0 ;
  treept [ nt ] = 0 ; 
  for( a = 2 ; a < nt ; ++ a ) { 
     n = a ; while ( bchbel [ n ] < 0 ) n = treept [ n ] ;
     while( ( jamp & bm [ bitz ] ) != jamp ) ++ bitz ;
     needbitz += ( bitz + 1 ) ; 
     if( ++ hrig [ n ] > 2 ) { 
         low = n ; 
         n = hlef [ n ] ;
         while ( bchbel [ n ] < 0 ) n = hsis [ n ] ;
         value = ( bchbel [ n ] << 1 ) | 1 ;
         n = treept [ a ] ;
         while( ! hrig [ n ] ) { ++ hrig [ n ] ; n = treept [ n ] ; }   
         n = a;
         ++ jamp; 
         while( n != low ) { bchbel [ n ] = jamp ; n = treept [ n ] ; } 
         bchbel [ a ] = jamp ; }
     else {
         at = bchbel [ n ] ;
         b = low = n ;
         while ( bchbel [ low ] == at ) low = treept [ low ] ;
         n = hlef [ n ] ; while ( bchbel [ n ] < 0 ) n = hsis [ n ] ;  
         value = bchbel [ n ] << 1 ; 
         n = treept [ a ] ; while ( ! hrig [ n ] ) { ++ hrig [ n ] ; n = treept [ n ] ; } 
         ++ jamp; 
         n = a ; while( n != b ) { bchbel [ n ] = jamp ; n = treept [ n ] ; } 
         ++ jamp;
         while( n != low ) { bchbel [ n ] = jamp ; n = treept [ n ] ; }}
     curkey |= ( value << ovr );
     ovr += bitz + 1 ; }
  curlokey =
      ( curkey & 65535 ) ^
      ( ( curkey & ( 65535 << 16 ) ) >> 16 ) ;
  return; 
}

void typify ( void )  // optimize and hash trees...
{
    char oplis[40] ;
    int a , b , c , stats , nod ;
    int bfor ; 
    double len ; 
    int nunos , shif ;
int foundit , numlclasses ;
double * lclass ; 
    FILE * frmpoo ; 
    Treetyp * entry ;
    frmpoo = fopen ( "allliks.tmp" , "r" ) ;
    if ( frmpoo == NULL ) {
        fprintf ( stderr , "OOPS!... can't find file from PAUP*\nMake sure PAUP* is in the same dir as likbias\n" ) ;
        exit ( 0 ) ; }
    oplis [ 0 ] = '\0' ;
    while ( strcmp ( oplis , "ambiguities" ) && strcmp ( oplis , "Tree" ) ) 
         fscanf ( frmpoo , " %s" , oplis ) ;
    if ( !strcmp ( oplis , "Tree" ) ) ungetc ( 'T', frmpoo ) ;
    lclass = malloc ( ntrees * sizeof ( double ) ) ;
    numlclasses = 0 ; 
    for ( a = 0 ; a < ntrees ; ++ a ) {
        if ( ( a % 127 ) == 0 ) 
           fprintf ( stderr , "\rchecking %i%%...              " , ( 100 * a ) / ntrees ) ; 
        anxpt = trees[a].tree ;
        reset_nodz ( anxpt ) ;
        binmksis ( anxpt ) ; 
        len = getdalik ( frmpoo ) ;
        foundit = 0 ; 
        for ( bfor = 0 ; bfor < numlclasses && !foundit ; ++ bfor ) {
            if ( lclass [ bfor ] < len && len - lclass [ bfor ] < TOLERANCE ) { len = lclass [ bfor ] ; foundit = 1 ; }
            if ( lclass [ bfor ] > len && lclass [ bfor ] - len < TOLERANCE ) { len = lclass [ bfor ] ; foundit = 1 ; }}
        if ( !foundit ) 
            lclass [ numlclasses ++ ] = len ; 
        if ( len < mptlen ) mptlen = len ; 
        if ( len > lptlen ) lptlen = len ; 
        trees[a].len = len ;
        hashtree ( anxpt ) ;
        trees[a].key = curkey ;
        trees[a].lokey = curlokey ; }
    for ( a = 0 ; a < ntrees ; ++ a )
       if ( trees[a].len == mptlen ) trees[a].isopt = 1 ;
       else trees[a].isopt = 0 ;
    for ( a = 0 ; a < 65536 ; ++ a ) treelist [ a ] = NULL ; 
    for ( a = 0 ; a < ntrees ; ++ a ) trees[a].to = NULL ;
    for ( a = 0 ; a < ntrees ; ++ a ) {
        shif = trees[a].lokey ; 
        entry = treelist [ shif ] ;
        if ( entry == NULL ) 
            treelist [ shif ] = trees + a ;
        else {
            while ( entry -> to != NULL ) 
                entry = ( Treetyp * ) entry -> to ;
            entry -> to = trees + a ; }}
    return ;         
}

void connect ( void )
{
    double perc ;
    int a , b ; 
    for ( a = b = 0 ; a < ntrees ; ++ a , ++ b ) {
        if ( b == 763 ) {
           perc = ( double ) a / ntrees ;
           perc *= 100 ;
           fprintf ( stderr , "\rConnecting trees... %3.2f%%" , perc ) ;
           b = 0 ; }
        interconnect ( a ) ; }
   fprintf ( stderr , "\rConnected %i trees                   \n" , ntrees ) ; 
}

void tbreroot ( int new, int old )
{
   char * an = anxpt ; 
   int i , j , k , l , m , n , nx , first ;
   if ( an [ new ] == old || old < nt ) return ; 
   j = nx = first = an [ i = new ] ;
   n = sis [ new ] ;
   while ( nx != old ) {
       k = i ; 
       i = j ;
       j = nx ;
       m = n ;
       n = sis [ j ] ; 
       if ( k == lef [ i ] ) { lef [ i ] = j ; sis [ sis [ rig [ i ] ] = j ] = rig [ i ] ; } 
       else { rig [ i ] = j ; sis [ sis [ lef [ i ] ] = j ] = lef [ i ] ; }
       nx = an [ j ] ; 
       an [ j ] = i ; }
   an [ lef [ j ] = m ] = an [ rig [ j ] = n ] = j ;
   sis [ sis [ n ] = m ] = n ;
   an [ lef [ old ] = new ] = an [ rig [ old ] = first ] = old ;
   sis [ sis [ new ] = first ] = new ; 
   return ;
}

void zert ( int whatp, int wherep )
{
    int a, b, c ;
    if ( wherep == 0 ) {
        anxpt [ lef [ nt ] = sis [ whatp ] = 0 ] = nt ; 
        anxpt [ rig [ nt ] = sis [ 0 ] = whatp ] = nt ;
        return ; } 
    a = anxpt [ wherep ] ;
    c = sis [ wherep ] ; 
    b = anxpt [ wherep ] = anxpt [ whatp ] ; 
    anxpt [ b ] = a;
    sis [ sis [ wherep ] = rig [ b ] = whatp ] = lef [ b ] = wherep;
    sis [ lef [ a ] = c ] = b ;
    sis [ rig [ a ] = b ] = c ;
}     

void unzert ( int whatp )
{  
    int a, b, c;
    a = anxpt [ whatp ] ;
    if ( a == nt ) {
        lef [ anxpt [ 0 ] = nt ] = rig [ nt ] = sis [ 0 ] = 0 ;
        anxpt [ whatp ] = whatp ;
        return ; } 
    b = sis [ whatp ] ;
    c = sis [ a ] ; 
    a = anxpt [ a ] ; 
    lef [ a ] = b ;
    rig [ anxpt [ b ] = anxpt [ c ] = a ] = sis [ sis [ c ] = b ] = c ;
}

int binoptlist ( char * list , char * tt )
{
    char * x , * y ;
    x = y = list ; 
    * y ++ = nt ;
    while ( x < y ) {
        if ( lef [ * x ] >= nt ) * y ++ = lef [ * x ] ;
        if ( rig [ * x ] >= nt ) * y ++ = rig [ * x ] ;
        ++ x ; }
    return y - list ;
}

void binmksis ( char * tt ) 
{
    int a , b , c , lownod = tt [ nt ] ;
    char lassis [ 20 ] ; 
    for ( a = nodz ; a -- ; ) nodfork [ a ] = 0 ;
    for ( a = nt ; a -- ; ) {
        b = tt [ c = a ] ;
        while ( b != lownod ) {
                 if ( !nodfork [ b ] ) 
                     lef [ b ] = lassis [ b ] = c ;
                 else 
                     sis [ sis [ c ] = lassis [ b ] ] = rig [ b ] = c ;
                 if ( ++ nodfork [ b ] > 1 ) break ;
                 c = b ;
                 b = tt [ b ] ; }}
    return ; 
}         

char downlis[20] ;

void interconnect ( int which )
{
    Treetyp * myt = trees + which ; 
    int nod ;
    double len , lim = myt -> len ;  
    int clip , bestpos , begpos , sisis , nunos ;
    int a , b , c , d , stats , curroot , tgt , rootbackto , rootelse ; 
    char biglist[20] , rlist[20] , tgtlist[20];
    char * rdo , * rdone , * tgtdo , * tgtdone ;
    if ( myt -> isopt ) return ; 
    anxpt = myt -> tree ;
    anxpt [ nt ] = nodz ;
    binmksis ( anxpt ) ;
    if ( !rig [ nt ] ) {
        rig [ nt ] = lef [ nt ] ;
        lef [ nt ] = 0 ; }
    nunos = binoptlist ( biglist , anxpt ) ;
    myt -> nmoves = 0 ;
    for ( a = 0 ; a < biz ; ++ a ) bitstring [ a ] = 0 ; 
    for ( clip = 1 ; clip < nodz ; ++ clip ) {
        if ( clip == nt ) continue ; 
        unzert ( clip ) ; 
        bestpos = sisis = sis [ clip ] ;
        len = 0 ;
        for ( a = nunos ; a -- ; ) {
            nod = biglist [ a ] ;
            if ( nod == anxpt [ clip ] ) continue ; 
            for ( c = 0 ; c < nc ; ++ c ) {
                stats = matrix [ lef [ nod ] ] [ c ] & matrix [ rig [ nod ] ] [ c ] ;
                if ( !stats ) {
                   stats = matrix [ lef [ nod ] ] [ c ] | matrix [ rig [ nod ] ] [ c ] ;
                   ++ len ; }
                matrix [ nod ] [ c ] = stats ; }}
        if ( len == lim ) {
            zert ( clip , bestpos ) ;
            continue ; }
        tgtdo = tgtdone = tgtlist ;
        if ( !sisis ) * tgtdo ++ = 0 ;
        else {
            * tgtdo ++ = sis [ 0 ] ;
            while ( tgtdone < tgtdo ) {
                nod = * tgtdone ++ ;
                if ( nod > nt ) {
                    * tgtdo ++ = lef [ nod ] ; 
                    * tgtdo ++ = rig [ nod ] ; }}}
        rdo = rdone = rlist ;
        if ( clip < nt ) * rdo ++ = clip ;
        else {
            * rdo ++ = rig [ clip ] ;
            rootbackto = lef [ clip ] ;
            rootelse = rig [ clip ] ; 
            if ( lef [ clip ] > nt ) {
                * rdo ++ = lef [ lef [ clip ] ] ;  
                * rdo ++ = rig [ lef [ clip ] ] ; }}
            if ( use_nni ) 
                if ( !sisis ) rdone = rdo ;
                else {
                   tgtdo = tgtdone = tgtlist ;
                   if ( sis [ sisis ] ) {
                       * tgtdo ++ = sis [ sisis ] ;
                       * tgtdo ++ = anxpt [ sisis ] ; }
                   else rdone = rdo ; }
             while ( rdone < rdo ) {
                 curroot = * rdone ;
                 if ( curroot > nt ) {
                    * rdo ++ = lef [ curroot ] ; 
                    * rdo ++ = rig [ curroot ] ; }
                 if ( clip > nt && rdone > rlist ) tbreroot ( curroot , clip ) ;
                 for ( tgtdone = tgtlist ; tgtdone < tgtdo ; ++ tgtdone ) {
                     tgt = * tgtdone ; 
                     zert ( clip , tgt ) ;
                     len = findmylik ( ) ; // dodalik ( ) ;
                     if ( len < lim ) 
                        addtolist ( myt , len ) ;
                     unzert ( clip ) ; }  // targets 
                 if ( use_spr || use_nni ) break ; 
                 ++ rdone ; }   // rootings 
        if ( clip > nt ) {
            if ( !use_spr && !use_nni ) tbreroot ( rootbackto , clip ) ;
            if ( lef [ clip ] != rootbackto ) {
                a = rig [ clip ] ;
                rig [ clip ] = lef [ clip ] ;
                lef [ clip ] = a ; }
            if ( lef [ clip ] != rootbackto || rig [ clip ] != rootelse )
                tbreroot ( rootelse , clip ) ; }
        zert ( clip , bestpos ) ; } // clippings 
    return ;
}

char tmptree[20] ;

double findmylik ( void )
{
    Treetyp * entry ;
    int a , whichis ;
    for ( a = 0 ; a < nodz ; ++ a ) tmptree [ a ] = anxpt [ a ] ;
    reset_nodz ( tmptree ) ; 
    hashtree ( tmptree ) ;
    entry = ( Treetyp * ) treelist [ curlokey ] ;
    while ( entry -> key != curkey )
        entry = ( Treetyp * ) entry -> to ;
    return entry -> len ;
}    

void addtolist ( Treetyp * myt , double len )
{
    Treetyp * entry ;
    int a , whichis ;
    for ( a = 0 ; a < nodz ; ++ a ) tmptree [ a ] = anxpt [ a ] ;
    reset_nodz ( tmptree ) ; 
    hashtree ( tmptree ) ;
    entry = ( Treetyp * ) treelist [ curlokey ] ;
    while ( entry -> key != curkey )
        entry = ( Treetyp * ) entry -> to ;
    whichis = entry - trees ; 
    if ( deldupes ) 
       if ( ( bitstring [ whichis / 32 ] & ( 1 << ( whichis % 32 ) ) ) ) return ; 
    if ( myt -> nmoves >= MAXSWAPS ) {
       fprintf ( stderr , "\nOOPS!!\nCannot store all swaps for tree %i!!!\n\nIncrease \"MAXSWAPS\", recompile, and try again!\n" , myt - trees ) ;
       exit ( 0 ) ; }
    myt -> tbrlist [ myt -> nmoves ++ ] = whichis ;
    bitstring [ whichis / 32 ] |= 1 << ( whichis % 32 ) ; 
}    

double treefac , divfac [ 50 ] ;   // paths should be MUCH shorter than this, but just in case...
int atlevel ;
int from [ 50 ] ;
Treetyp * curtree ; 

void savant ( what )
{
    int a ;
    fprintf ( stdout , "(" ) ;
    for ( a = 0 ; a < nodz ; ++ a )
      if ( anxpt [ a ] == what )
        if ( a >= nt ) savant ( a ) ;
        else fprintf ( stdout , "%c " , taxname[a] ) ;
    fprintf ( stdout , ")" ) ;
}

void do_user_groups ( void )
{
    double totfreq[25] , optfreq[25] , suboptfreq[25] , expecfreq[25] ;
    int a , i , n , j , k , isamatch , nod , nunos ;
    char list[20] ;
    double eqfactor ;
    unsigned long int myinner , myouter ; 
    Treetyp * myt ; 
    for ( a = 0 ; a < nt ; ++ a )
       testmat [ a ] = 1 << a ;
    if ( numtestpartits ) {
        for ( a = 0 ; a < numtestpartits ; ++ a ) 
            totfreq [ a ] = expecfreq [ a ] = 0 ;
        eqfactor = ( double ) 1 / nummpts ; 
        for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
            if ( myt -> nmoves ) continue ;
            anxpt = myt -> tree ;
            binmksis ( anxpt ) ;
            nunos = binoptlist ( list , anxpt ) ;
            for ( i = nunos ; i -- ; ) {
                nod = list [ i ] ;
                testmat [ nod ] = testmat [ lef [ nod ] ] | testmat [ rig [ nod ] ] ; }
            for ( n = 0 ; n < numtestpartits ; ++ n ) {
                isamatch = 0 ;
                myinner = myouter = 0 ;
                for ( k = 0 ; k < partitsize [ n ] ; ++ k )
                  if ( k < partitat [ n ] ) 
                      myouter |= ( 1 << testpartit [ n ] [ k ] ) ;
                  else 
                      myinner |= ( 1 << testpartit [ n ] [ k ] ) ;
                for ( i = nunos ; i -- && !isamatch ; ) {
                    nod = list [ i ] ;
                    if ( ( myinner & testmat [ nod ] ) == myinner && !( testmat [ nod ] & myouter ) ) isamatch = 1 ; 
                    if ( ( myouter & testmat [ nod ] ) == myouter && !( testmat [ nod ] & myinner ) ) isamatch = 1 ; }
                if ( isamatch ) {
                    totfreq [ n ] += myt -> prob ;
                    if ( myt -> isopt ) 
                        expecfreq [ n ] += eqfactor ; }}}
        fprintf ( stdout , "\nPARTITION SUMMARY: \n\n       partition       TOTAL         expected\n\n   " ) ;
        for ( a = 0 ; a < numtestpartits ; ++ a ) {
            fprintf ( stdout , "    " ) ; 
            for ( k = 0 ; k < partitsize [ a ] ; ++ k ) {
                if ( k == partitat [ a ] ) 
                     fprintf ( stdout , "|"  ) ; 
                fprintf ( stdout , "%c" , taxname [ testpartit [ a ] [ k ] ] ) ; }
            for ( k = partitsize [ a ] ; k < 10 ; ++ k ) fprintf ( stdout , " " ) ;     
            fprintf ( stdout , "     %2.8f    %2.8f\n   " , totfreq [ a ] , expecfreq [ a ] ) ; }
        fprintf ( stdout , "\n" ) ; }
    if ( !numtestgroups ) return ;
    for ( a = 0 ; a < numtestgroups ; ++ a ) 
        totfreq [ a ] = optfreq [ a ] = suboptfreq [ a ] = expecfreq [ a ] = 0 ;
    eqfactor = ( double ) 1 / nummpts ; 
    for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
        if ( myt -> nmoves ) continue ;
        anxpt = myt -> tree ;
        binmksis ( anxpt ) ;
        nunos = binoptlist ( list , anxpt ) ;
        for ( i = nunos ; i -- ; ) {
            nod = list [ i ] ;
            testmat [ nod ] = testmat [ lef [ nod ] ] | testmat [ rig [ nod ] ] ; }
        for ( n = 0 ; n < numtestgroups ; ++ n ) {
            isamatch = 0 ; 
            for ( j = nt + 1 ; j < nodz && !isamatch ; ++ j )
                if ( testmat [ j ] == testgroup [ n ] ) isamatch = 1 ; 
            if ( isamatch ) {
                totfreq [ n ] += myt -> prob ;
                if ( myt -> isopt ) {
                    optfreq [ n ] += myt -> prob ;
                    expecfreq [ n ] += eqfactor ; }
                else suboptfreq [ n ] += myt -> prob ; }}}
    fprintf ( stdout , "\nGROUP SUMMARY: \n\n         group     optimal       subopt        TOTAL         expected\n\n   " , numtestgroups ) ;
    for ( a = 0 ; a < numtestgroups ; ++ a ) {
        for ( j = testgroupsz[ a ] ; j < 10 ; ++ j ) fprintf ( stdout , " " ) ; 
        for ( i = 0 ; i < nt ; ++ i ) 
             if ( ( ( 1 << i ) & testgroup [ a ] ) )
                 fprintf ( stdout , "%c" , toupper ( taxname[i] ) ) ; 
        fprintf ( stdout , "      %2.8f    %2.8f    %2.8f    %2.8f\n   " ,
             optfreq[ a ] , suboptfreq[a] , totfreq[a] , expecfreq[a] ) ; }
    fprintf ( stdout , "\n" ) ; 
}

void curseshow ( Treetyp * tp )
{
    int a , j ;
    double thisprob ; 
    Treetyp * goesto ; 
    divfac [ atlevel ] = ( double ) 1 / tp -> nmoves ;
    for ( a = 0 ; a < atlevel ; ++ a ) fprintf ( stdout , "    " ) ;
    anxpt = tp -> tree ;
    savant ( nt ) ;
    fprintf ( stdout , "  [%i]  - L %f " , tp - trees , tp -> len ) ; 
    if ( tp -> nmoves == 0 ) {
        fprintf ( stdout , " -  P = %.8f" , treefac ) ;
        if ( !tp -> isopt ) fprintf ( stdout , " !! " ) ; }
    fprintf ( stdout , "\n\n" ) ; 
    ++ atlevel ; 
    for ( a = 0 ; a < tp -> nmoves ; ++ a ) {
        goesto = trees + ( tp -> tbrlist [ a ] ) ;
        if ( !goesto -> nmoves ) {
            thisprob = 1 ;
            for ( j = 0 ; j < atlevel ; ++ j )
                 thisprob *= divfac [ j ] ;
            goesto -> prob += thisprob ;
            for ( j = 0 ; j < atlevel ; ++ j ) fprintf ( stdout , "    " ) ;
            anxpt = goesto -> tree ; 
            savant ( nt ) ;
            fprintf ( stdout , "  [%i]  - L %f " , goesto - trees , goesto -> len ) ; 
            fprintf ( stdout , " -  P = %.8f" , thisprob ) ;
            if ( !goesto -> isopt ) fprintf ( stdout , " !! " ) ; 
            fprintf ( stdout , "\n\n" ) ; }
        else curseshow ( goesto ) ; }
    -- atlevel ;
}    

void showpaths ( void )
{
    int a , b ;
    Treetyp * myt ;
    double perc , probsum = 0 ;
    treefac = ( double ) 1 / ntrees ; 
    myt = trees + pathsfrom ;
    atlevel = 0 ;
    fprintf ( stdout , "\n\nPATHS FROM TREE %i\n\n" , pathsfrom ) ; 
    curseshow ( myt ) ;
    fprintf ( stdout , "\n" ) ; 
}

void calculate_tree_probs ( void )
{
    Treetyp * myt ;
    int a , b , indx ;
    int * damoves , trees_to_count ;
    double curfac , probsum = 0 ;
    double optminprob = 1000 , optmaxprob = 0 ;
    double nextlen = lptlen , curlen ; 
    Treetyp * to ;
    treefac = ( double ) 1 / ntrees ;
    for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
        myt -> prob = treefac ; 
        myt -> to = NULL ; }
    nummpts = numnonmpts = 0 ;
    trees_to_count = ntrees ;
    while ( trees_to_count ) {
        fprintf ( stderr , "\rCalculating probs %.0f%%\r" , ( double ) ( ntrees - trees_to_count ) / ntrees ) ; 
        curlen = nextlen ;
        nextlen = 0 ; 
        for ( a = 0 , to = trees ; a < ntrees ; ++ a , ++ to ) {
            if ( to -> len > curlen ) continue ;
            if ( to -> len < curlen ) {
                if ( to -> len > nextlen ) nextlen = to -> len ;
                continue ; }
            -- trees_to_count ; 
            if ( ! to -> nmoves ) continue ;
            curfac = to -> prob / to -> nmoves ;
            for ( b = 0 , damoves = to -> tbrlist ; b < to -> nmoves ; ++ b , ++ damoves ) 
                trees  [ * damoves ] . prob += curfac ; }}
    // show results...
    fprintf ( stderr , "                                       \r" ) ; 
    fprintf ( stdout , "\n" ) ; 
    for ( a = 0 , myt = trees ; a < ntrees ; ++ a , ++ myt ) {
        if ( myt -> nmoves ) continue ;
        if ( myt -> isopt ) {
            fprintf ( stdout , "TREE %6i - %.8f - L = %f " , a , myt -> prob , myt -> len ) ;
            ++ nummpts ; }
        else {
            fprintf ( stdout , "tree %6i - %.8f - L = %f " , a , myt -> prob , myt -> len ) ;
            ++ numnonmpts ; }
        probsum += myt -> prob ;
        if ( myt -> prob > damaxprob ) damaxprob = myt -> prob ; 
        if ( myt -> prob < daminprob ) daminprob = myt -> prob ;
        if ( myt -> isopt ) {
           if ( myt -> prob > optmaxprob ) optmaxprob = myt -> prob ; 
           if ( myt -> prob < optminprob ) optminprob = myt -> prob ; }
        /*** show it ***/
        anxpt = myt -> tree ;
        savant ( nt ) ;
        fprintf ( stdout , ";\n" ) ; }
   fprintf ( stdout , "\nTOTAL PROBS.: %.8f (%i MLT's, %i suboptrees)\n" , probsum , nummpts , numnonmpts ) ;
   curfac = damaxprob / daminprob ;
   if ( damaxprob != daminprob ) 
       fprintf ( stdout , "\nMin-Max probs.: %.8f - %.8f (Ratio %.3f)\n" , daminprob , damaxprob , curfac ) ; 
   if ( optminprob != optmaxprob && numnonmpts ) {
       curfac = optmaxprob / optminprob ;
       fprintf ( stdout , "\nMin-Max probs. on OPTIMAL trees: %.8f - %.8f (Ratio %.3f)\n" , optminprob , optmaxprob , curfac ) ; }
   fprintf ( stdout , "\n" ) ; 
}
                
void main ( int argc , char ** argv )
{
    int a ;
    char * fnameis ;
    if ( argc < 2 ) {
        fprintf ( stderr , 
        "\n"
        "  LIKBIAS\n"
        "\n"
        "  calculates (by enumeration) probabilities of end-points \n"
        "  of SPR/TBR trees (on a single tree), starting from random\n"
        "  trees, for fully randomized SPR/TBR.\n"
        "\n"
        "  usage:\n"
        "    likbias [spr tbr rspr rtbr] datafile [ab] [cd] [path N] [ab/cd]\n"
        "\n"
        "  datafile is in xread format, but without xread itself; taxon\n"
        "  names must be single letters. Only states 0-4 allowed.\n"
        "  If datafile name is preceded by an asterisk, then PAUP* is not\n"
        "  called to calculate likelihoods (using last files generated).\n"
        "\n"
        "  \"spr\" or \"tbr\" do not use redundant moves; \"rspr\" and \"rtbr\" do.\n"
        "\n"
        "  \"ab\", \"cd\", etc., are groups to test (probability is the\n"
        "  sum of probabilities of all SPR or TBR end-points with the group)\n"
        "\n"
        "  \"path N\" shows the paths from tree nr. N to all end-points (with \n"
        "  conditional probabilities shown; note in this case initial tree is \n"
        "  given, with P=1, instead of being P = 1/ntrees)\n" 
        "\n"
        "\n\"ab/cd\" checks the probability of a quartet to be resolved as AB|CD\n"
        "\n"
        "  You can redirect output with \"> outfile\"\n"
        "\n"
         ) ; 
        exit ( 0 ) ; }
    fprintf ( stdout , "\n\nOptions: " ) ; 
    for ( a = 1 ; a < argc ; ++ a ) fprintf ( stdout , "%s " , argv[a] ) ;
    fprintf ( stdout , "\n" ) ;
    fflush ( stdout ) ; 
    a = 1 ; 
    if ( !strcmp ( argv[1] , "spr" ) || !strcmp ( argv[1] , "rspr" ) ) {
        use_spr = 1 ;
        if ( argv[1][0] == 'r' ) deldupes = 0 ;
        a = 2 ; }
    else if ( !strcmp ( argv[1] , "nni" ) ) {
        use_nni = deldupes = 1 ;
        a = 2 ; }
    else if ( !strcmp ( argv[1] , "tbr" ) || !strcmp ( argv[1] , "rtbr" ) ) {
        use_spr = 0 ;
        if ( argv[1][0] == 'r' ) deldupes = 0 ;
        a = 2 ; }
    if ( a >= argc ) {
        fprintf ( stderr , "No data!" ) ;
        exit ( 0 ) ; }
    readdata ( argv [ a ] ) ;
    if ( a + 1 < argc ) 
        if ( !strcmp ( argv[ a + 1 ] , "path" ) ) {
            pathsfrom = atoi ( argv [ a + 2 ] ) ;
            fprintf ( stderr , "\nShowing path(s) from tree %i\n" , pathsfrom ) ; }
        else 
          while ( ++ a < argc ) 
            addtestgroup ( argv [ a ] ) ; 
    generalltrees ( ) ;
    typify ( ) ;
    connect ( ) ;
    if ( pathsfrom >= 0 )
       showpaths () ;
    else {
       calculate_tree_probs ( ) ; 
       do_user_groups ( ) ; }
    exit ( 1 ) ; 
}


double getdalik ( FILE * in )
{
    float val ;
    char x = ' ' ;
    while ( isspace ( x ) ) x = getc ( in ) ;
    if ( x == 'T' ) 
        while ( x != 'L' ) x = getc ( in ) ;
    else ungetc ( x , in ) ;
    fscanf ( in , " %f" , &val ) ; 
    return ( double ) val ; 
}    


FILE * nxfile ; 
int transtable[20] ;
char cmdname[50] ; 

char gtnb ( void )
{
    int x = getc ( nxfile ) ;
    while ( isspace ( x ) ) x = getc ( nxfile ) ;
    return x ; 
}

char peep ( void )
{
    int x = getc ( nxfile ) ;
    while ( isspace ( x ) ) x = getc ( nxfile ) ;
    ungetc ( x , nxfile ) ; 
    return x ; 
}

void read_single_tree ( void )
{
    int htu, unbalpars, atnode, i , car ;
    char * an = trees [ ntrees ].tree ;
    int x = ' ' ;
    while ( x != ']' ) x = getc ( nxfile ) ; 
    car = gtnb () ; 
    for( htu = nt; htu--; ) an[htu] = 0; 
    htu = atnode = nt - 1; 
    unbalpars = 0; 
    while(car!=';') { 
       if(car=='(')
          { an[htu+1]=atnode;
            atnode = ++htu;
            --unbalpars; } 
      else  
        { if(car==')') { atnode = an[atnode]; ++unbalpars; }
          else {
                if ( isdigit ( car ) ) {
                     ungetc ( car , nxfile ) ; 
                     fscanf ( nxfile , " %i" , &i ) ;
                     if ( i < 1 || i > nt ) {
                         fprintf ( stderr , "No taxon %i\nPossibly mistaken file!\n" , i ) ;
                         exit ( 0 ) ; }
                     i = transtable[i] ; }
                an[ i ] = atnode; }
          if ( ( peep () ) == ',' ) getc ( nxfile ) ; }
      car = gtnb(); }
   an[nt] = ++htu ;
   ntrees ++ ; 
   return ;
} 

void read_all_trees ( void )
{
    int x ;
    int a ; 
    int sendto ;
    int tord = ntrees ; 
    nxfile = fopen ( "poofile.tre" , "r" ) ;
    if ( nxfile == NULL ) {
        fprintf ( stderr , "\nError: can't open tree file\n" ) ;
        exit ( 0 ) ; }
    for ( a = nt ; a -- ; ) transtable[a+1] = a ;
    x = ' ' ;
    while ( x != ';' ) x = getc ( nxfile ) ;
    x = ' ' ;
    while ( x != ';' ) x = getc ( nxfile ) ;
    strcpy ( cmdname , "      " ) ;
    ntrees = 0 ; 
    while ( strcmp ( cmdname , "End;" ) && strlen ( cmdname ) > 2 ) {
            while ( x == ';' ) { x = getc ( nxfile ) ; x = peep () ; }
            fscanf ( nxfile , " %s" , cmdname ) ;
            if ( !strcmp ( cmdname , "tree" ) ) read_single_tree () ;
            else {
                if ( tord == ntrees ) break ; 
                x = ' ' ;
                while ( x != ';' ) x = getc ( nxfile ) ; }
            x = peep () ; }
     fclose ( nxfile ) ; 
     return ; 
}
    



    
    
